/*

  This file contains procedures to solve a Linear Programming problem
  with inequality and equality constraints.  by the dual simplex
  method.
  
   - The code was originally in Algol 60 and was published in
   Collected Algorithms from CACM (alg #333) by Rudolfo C. Salazar and
   Subrata K. Sen in 1968 under the title MINIT (MINimum ITerations)
   algorithm for Linear Programming.
   
   - It was directly translated into C by Badri Lokanathan, Dept. of
   EE, University of Rochester, with no modification to code
   structure.

   - The spaghetti was chopped into edible reentrant bits in ANSI C by Han-Wen
   Nienhuys <hanwen@cs.uu.nl> */


/*
   
   The problem statement is

   Maximise z = cX
  
   subject to
   AX <= b
   X >=0
  

*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "lp.h"
#include "xmalloc.h"
#include "tableau.h"


#define ABS(A) ((A) > 0 ? (A): -(A))
#define MAX(A,B) ((A) > (B) ? (A) : (B))
#define MIN(A,B) ((A) < (B) ? (A) : (B))




/*
  Parameters (globals)
     
   MINIT_TOLERANCE: nonnegativity minit_tolerance.  In compares, numbers bigger
   than -MINIT_TOLERANCE are considered non-negative (default 1e-8)
   These compares are done to check the optimality of a solution.

   The value of MINIT_TOLERANCE should reflect the magnitude of the
   co-efficient matrix.


   
   DEBUG_MINIT: print gobs of debug info
*/

Real LARGENUMBER =1e6;
Real VLARGENUMBER= 1e8;
Real VSMALLNUMBER =1e-8;
Real minit_tolerance=1e-8;
int debug_minit;

void tune_minit (Real infty, Real inffty, Real epsilon, Real tolerance)
{
  LARGENUMBER = infty;
  VLARGENUMBER = inffty;
  VSMALLNUMBER = epsilon;
  minit_tolerance = tolerance;
}


void determine_pivot_columns(struct TABLEAU *tabp);
void determine_pivot_rows(struct TABLEAU *tabp);
int minimum_iterations_lp_solve(struct TABLEAU *tabp, struct LP_SOLUTION *);
int phase1(struct TABLEAU *tabp);
int do_pivot(struct TABLEAU *tabp, int p, int q);

int
lp_solve(struct LP_PROBLEM *problem_p, struct LP_SOLUTION *lps)
{
  static struct TABLEAU * tabp;
  int err;

  if (!tabp)
    {
      tabp = new_tableau();
    }
  
  set_tableau_for_problem (tabp, problem_p);
  err=  minimum_iterations_lp_solve(tabp, lps);
  /*  free__TABLEAU(tabp);*/
  return err;
}

char *errmsg[] = 
{
 "everything went OK",
 "no solution existed."
 "primal had no feasible solutions",
 "primal had unbounded solutions"
};


char const *
lp_solve_error (int i)
{
  assert (i >= 0 && i < 4);
  return errmsg [i];
}

static void
find_negative_coefs (struct TABLEAU *tabp)
{
  int i,j;

  tabp->no_neg_costcoef = 1;	/* UGH: array starts at 1 */
  tabp->no_neg_rhs = 1;		/* UGH: idem. */
  
  for(j = 0; j < tabp->columns - 1; j++)
    if(tabp->elements[0][j] < -minit_tolerance)
      tabp->neg_costcoef_indices[(tabp->no_neg_costcoef++)-1] = j;
    
  for(i = 1; i < tabp->rows; i++)
    if(tabp->elements[i][tabp->columns-1] < -minit_tolerance)
      tabp->neg_rhs[(tabp->no_neg_rhs++)-1] = i;
}

/*
   This is the main procedure. 

   PARAMETERS

   TABP: the initialised tableau

   SOLP: prealloced solution


   RETURN
    
   0 if everything went OK and x, w, z as the solutions.

   1 if no solution existed.
   
   2 if primal had no feasible solutions.

   3 if primal had unbounded solutions.

*/

int
minimum_iterations_lp_solve(struct TABLEAU *tabp, struct LP_SOLUTION *solp)
{
  register int i, j;
  
  /* Now begins the actual algorithm. */
  if(tabp->no_eq)
    {
      if(phase1(tabp))
	return 1;
    }

  while (1) {
    if(debug_minit)
      print_tableau(tabp);	/* Print the tableau. */

    find_negative_coefs (tabp);
    if(tabp->no_neg_costcoef == 1)
      {
	if(tabp->no_neg_rhs == 1) /* Optimality: exit loop */
	  break;			/* EXIT LOOP! */

	if(tabp->no_neg_rhs == 2)
	  {
	    int infeasible = 1;
		
	    for(j = 0; infeasible && j < tabp->columns - 1; j++)
	      if(tabp->elements[tabp->neg_rhs[0]][j] < 0)
		infeasible = 0;

	    /* Primal problem has no feasible solutions. */
	    if (infeasible)
	      return(2);
	  }
	    
	determine_pivot_rows(tabp);
	if(do_pivot(tabp, tabp->imax,tabp->jm))
	  return(1);
      }
    else
      {
	if(tabp->no_neg_costcoef == 2)
	  if(tabp->no_neg_rhs == 1)
	    {
	      int unbounded = 1;
	      for(i = 1; unbounded && i < tabp->rows; i++)
		{
		  if(tabp->elements[i][tabp->neg_costcoef_indices[0]] > 0)
		    unbounded = 0;
		}

	      /* Primal problem is unbounded. */
	      if (unbounded)
		return(3);
	    }

	if(tabp->no_neg_rhs == 1)
	  {
	    determine_pivot_columns(tabp);
	    if(do_pivot(tabp, tabp->im,tabp->jmin))
	      return(1);
	  }
	else
	  {
	    determine_pivot_columns(tabp);
	    determine_pivot_rows(tabp);

	    if(tabp->gmin == LARGENUMBER)
	      {
		if(do_pivot(tabp,tabp->imax,tabp->jm))
		  return(1);
	      }
	    else if(tabp->phimax == - LARGENUMBER)
	      {
		if(do_pivot(tabp,tabp->im,tabp->jmin))
		  return(1);

	      }
	    else
	      {
		if(ABS(tabp->phimax) > ABS(tabp->gmin))
		  {
		    if(do_pivot(tabp,tabp->imax,tabp->jm))
		      return(1);
		  }
		else
		  {
		    if(do_pivot(tabp, tabp->im,tabp->jmin))
		      return(1);
		  }
	      }
	  }
      }
  }

  tableau_read_solution (tabp, solp);
  return(0);
}


/*

  Pivot around (p,q)

  IN
  
  P, Q: pivot

  IN, OUT

  TABP: the tableau

  RETURN:
  
   0: Everything was OK.
   1: No solution.
*/

int
do_pivot(struct TABLEAU *tabp, int p, int q)
{
  register int i, j;
  Real dummy;

  if(p == -1 || q == -1) /* No solution. */
    return(1);

  dummy = tabp->elements[p][q];

  if(debug_minit)
    printf("--\nPivot element is tabp->elements[%d][%d] = %f\n",p,q,dummy);

  for(j = 0; j < tabp->columns; j++)
    tabp->elements[p][j] /= dummy; 

  for(i = 0; i < tabp->no_cons + 1; i++)
    {
      if(i != p && tabp->elements[i][q] != 0.0)
	{
	  dummy = tabp->elements[i][q];
	  for(j = 0; j < tabp->columns; j++)
	    tabp->elements[i][j] -= tabp->elements[p][j] * dummy;
	}
    }

  tabp->chk[p] = q;
  return(0);
} /* do_pivot */


/*
   Performs calculations over columns to determine the pivot element.
*/
void determine_pivot_columns(struct TABLEAU *tabp)
{
  Real theta, gamma;
  register int i, L1;
  
  tabp->gmin = LARGENUMBER;
  tabp->jmin = -1;		/* Array tabp->indices in C start from 0 */

  for(L1 = 0; L1 < tabp->no_neg_costcoef - 1; L1++) /* huh? Why -1? --hwn */
    {
      int ncc_idx = tabp->neg_costcoef_indices[L1];
      
      tabp->imin[ncc_idx] = 0;
      tabp->thmin[ncc_idx] = LARGENUMBER;
      for(i = 1; i < tabp->no_cons + 1; i++)
	{
	  if(tabp->elements[i][ncc_idx] > minit_tolerance
	     && tabp->elements[i][tabp->columns-1] >= -minit_tolerance)
	    {
	      theta = tabp->elements[i][tabp->columns-1] / tabp->elements[i][ncc_idx];
	      if(theta < tabp->thmin[ncc_idx])
		{
		  tabp->thmin[ncc_idx] = theta;
		  tabp->imin[ncc_idx] = i;
		}
	    }
	}

      if(tabp->thmin[ncc_idx] == LARGENUMBER)
	gamma = VLARGENUMBER;
      else
	gamma = tabp->thmin[ncc_idx] * tabp->elements[0][ncc_idx];

      if(gamma < tabp->gmin)
	{
	  tabp->gmin = gamma;
	  tabp->jmin = ncc_idx;
	}
    }
  if(tabp->jmin > -1)
    tabp->im = tabp->imin[tabp->jmin];
} /* determine_pivot_columns */







/*
   Performs calculations over rows to determine the pivot element.
*/
void determine_pivot_rows(struct TABLEAU *tabp)
{
  Real delta, phi;
  register int j, k1;

  tabp->phimax = - LARGENUMBER;
  tabp->imax = -1;		/* Array indices in C start from 0 */
	
  for(k1 = 0; k1 < tabp->no_neg_rhs - 1; k1++)
    {
      int nr_idx = tabp->neg_rhs[k1];
      tabp->jmax[nr_idx] = 0;
      tabp->delmax[nr_idx] = - LARGENUMBER;
      for(j = 0; j < tabp->columns - 1; j++)
	{
	  if(tabp->elements[nr_idx][j] < -minit_tolerance && tabp->elements[0][j] >= -minit_tolerance)
	    {
	      delta = tabp->elements[0][j] / tabp->elements[nr_idx][j];
	      if(delta > tabp->delmax[nr_idx])
		{
		  tabp->delmax[nr_idx] = delta;
		  tabp->jmax[nr_idx] = j;
		}
	    }
	}

      if(tabp->delmax[nr_idx] == - LARGENUMBER)
	phi = - VLARGENUMBER;
      else
	phi = tabp->delmax[nr_idx] * tabp->elements[nr_idx][tabp->columns-1];
		
      if(phi > tabp->phimax)
	{
	  tabp->phimax = phi;
	  tabp->imax = nr_idx;
	}
    }
  if(tabp->imax > -1)
    tabp->jm = tabp->jmax[tabp->imax];
} /* determine_pivot_rows */




/*
  Find initial basic solution.
*/
int phase1(struct TABLEAU *tabp)
{
  Real theta, gamma;
  register int i, j, r, s;

  /* Fix suggested by Holmgren, Obradovic, Kolm. */
  register int im1, jmin1, first;

  im1 = jmin1 = -1;

  
  /* Fix suggested by Messham to allow negative coeffs. in
    equality constraints.
   */
  
  for(i = tabp->no_ineq + 1; i < tabp->no_cons + 1; i++)
    {
      if(tabp->elements[i][tabp->columns - 1] < 0)
	{
	  for(j = 0; j < tabp->columns; j++)
	    tabp->elements[i][j] = -tabp->elements[i][j];
	}
    }

  for(r = 0; r < tabp->no_eq; r++)
    {
      tabp->gmin = LARGENUMBER;
      tabp->no_neg_costcoef = 1;
      tabp->jmin = -1;
      first = 1;
      for(j = 0; j < tabp->no_vars; j++)
	{
	  tabp->thmin[j] = LARGENUMBER;
	  /* Fix suggested by Kolm and Dahlstrand */
	  /* if(tabp->elements[0,j] < 0) */
	  if(tabp->elements[0][j] < -minit_tolerance)
	    tabp->neg_costcoef_indices[(tabp->no_neg_costcoef++)-1] = j;
	}

    I_AM_A_GOTOLABEL_SHOOT_ME:				/* ugh */


      if(tabp->no_neg_costcoef == 1)
	{
	  for(j = 0; j < tabp->no_vars; j++)
	    tabp->neg_costcoef_indices[j] = j;
	  
	tabp->no_neg_costcoef = tabp->no_vars + 1;
	}

      for(s = 0; s < tabp->no_neg_costcoef - 1; s++)      
	{
	  int nc_idx =  tabp->neg_costcoef_indices[s];
	  for(i = tabp->no_ineq + 1; i < tabp->no_cons + 1; i++)
	    if(tabp->chk[i] == -1)
	      {
		/* Fix suggested by Kolm
		 * and Dahlstrand
		 */

		if(tabp->elements[i][nc_idx] > minit_tolerance)
		  {
		    if((theta = tabp->elements[i][tabp->columns-1] /
			tabp->elements[i][nc_idx]) < tabp->thmin[nc_idx])
		      {
			tabp->thmin[nc_idx] = theta;
			tabp->imin[nc_idx] = i;
		      }
		  }
	      }

	  /* Fix suggested by Obradovic overrides
	   * fixes suggested by Kolm and Dahstrand
	   * as well as Messham.
	   */
	  if(tabp->thmin[nc_idx] < LARGENUMBER)
	    {
	      gamma = tabp->thmin[nc_idx] * tabp->elements[0][nc_idx];
	      if(gamma < tabp->gmin)
		{
		  tabp->gmin = gamma;
		  tabp->jmin = nc_idx;
		}
	    }
	}
      if(tabp->jmin == -1)
	{
	  if(first)
	    {
	      first = 0;
	      tabp->no_neg_costcoef = 1;
	      goto I_AM_A_GOTOLABEL_SHOOT_ME;
	    }
	  else
	    tabp->im = -1;
	}
      else
	tabp->im = tabp->imin[tabp->jmin];

      if(tabp->im == im1 && tabp->jmin == jmin1)
	{
	  tabp->no_neg_costcoef = 1;
	  goto I_AM_A_GOTOLABEL_SHOOT_ME;
	}

      if(debug_minit)
	print_tableau(tabp);	/* Print the tableau. */

      if(do_pivot(tabp, tabp->im,tabp->jmin))
	return(1);
		
      im1 = tabp->im;
      jmin1 = tabp->jmin;
    }
  return(0);
} /* phase1 */


